"""
I find that pyplot is a pain to use and more importantly one often needs to interact with it
inbetween other code.

This intends to provide flexible enough wrappers around pyplots API,
that can be used in a more declarative way.
This makes it easier to differentiate between whats plotting code,
and what is unrelated, and also from the plotting code,
what makes it in the resulting pdf. 
"""

import matplotlib.pyplot as plt
from typing import Tuple, Literal, Iterable, Callable, NamedTuple, Callable
from abc import ABC, abstractmethod
from dataclasses import dataclass
import dataclasses
from .utils import UNREACHABLE
import numpy.typing as npt
import numpy as np

@dataclass
class AxData:
    label: str | None = None
    bottom: float | None = None
    top: float | None = None
    scale: Literal["log"] | Literal["linear"] = "linear"

class Plottable(ABC):
    @abstractmethod
    def apply(self, ax: plt.Axes):
        UNREACHABLE("please implement")

@dataclass
class Plot(Plottable):
    yData: Iterable
    xData: Iterable | None = None
    kwargs: dict = dataclasses.field(default_factory=lambda: {})
    label: str | None = None
    kind: Literal["line"] | Literal["scatter"] = "line"

    def apply(self, ax: plt.Axes):
        xData = self.xData if self.xData else range(len(self.yData))
        match self.kind:
            case "line": pass
            case "scatter": self.kwargs = {"marker": "x", "linestyle": "None", **self.kwargs}
            case _: UNREACHABLE("unrecognized plot kind")
        ax.plot(xData, self.yData, label = self.label, **self.kwargs)

@dataclass
class Bar(Plottable):
    heights: Iterable
    xPositions: Iterable | None = None
    width: float | None = None
    kwargs: dict | None = None
    label: str | None = None

    def apply(self, ax: plt.Axes):
        xPositions = self.xPositions if self.xPositions is not None else range(len(self.heights))
        ax.bar(xPositions, self.heights, label = self.label, width=self.width, **(self.kwargs if self.kwargs else {}))

@dataclass
class Function(Plottable):
    function: Callable[[float], float]
    "The function to plot, takes x value, returns y value"
    range_from: Iterable[npt.NDArray[float]]
    samples: int = 2
    kwargs: dict | None = None
    label: str | None = None

    def apply(self, ax: plt.Axes):
        x_max = max(map(lambda range_series: np.max(range_series), self.range_from))
        x_min = min(map(lambda range_series: np.min(range_series), self.range_from))
        x_vals = np.linspace(x_min, x_max, self.samples)
        ax.plot(x_vals, self.function(x_vals), label = self.label , **(self.kwargs if self.kwargs else {}))

@dataclass
class VertLine(Plottable):
    position: float
    linestyle: str
    color: str = "green"
    kwargs: dict | None = None
    label: str | None = None

    def apply(self, ax: plt.Axes):
        ax.axvline(self.position, label = self.label, linestyle = self.linestyle , **(self.kwargs if self.kwargs else {}))

@dataclass
class InMemoryPlot:
    fig: plt.Figure
    ax: plt.Axes
    save_path: str | None = None

    def draw_and_save(self):
        assert self.save_path != None
        self.fig.savefig(self.save_path)
        plt.close(self.fig)

@dataclass(frozen=False)
class PlotData:
    # xdata: 
    graphs: list[Plottable]
    x_ax: AxData
    y_ax: AxData
    title: str | None = None
    save_path: str | None = None

    def create(self, add_legend: bool = True) -> InMemoryPlot:
        fig: plt.Figure = plt.figure()
        ax: plt.Axes = fig.add_subplot(111)
        if self.title:
            ax.set_title(self.title)
        for graph in self.graphs:
            graph.apply(ax)
        ax.set_yscale(self.y_ax.scale)
        ax.set_xscale(self.x_ax.scale)
        if self.y_ax.label:
            ax.set_ylabel(self.y_ax.label)
        if self.x_ax.label:
            ax.set_xlabel(self.x_ax.label)
        ax.set_ylim(self.y_ax.bottom, self.y_ax.top)
        ax.set_xlim(self.x_ax.bottom, self.x_ax.top)
        if add_legend:
            ax.legend()
        ax.grid(True)
        fig.tight_layout()
        return InMemoryPlot(fig, ax, self.save_path)
