# ENEM aggregation code

This code does most of the grunt work of aggregating the massive interaction files from INEP for ENEM, as well as aggregating the metafiles also provided by INEP for each year (which contain their IRT values for the questions), and also combining these with the question and answer text provided by Marinho.

It does *not* necessarily put the results into the concrete format used by the models. That is done in Python to allow for more model-specific customization in the model repository itself.

Originally, this code was intended to run in a streaming fashion on any PC.
However, the streaming did not work out, so part of it is now done on the cluster.
Since there was already a lot of code when we gave up we kept it around, as I can not spare the time to port it to Python.

This was done in Rust mainly for 3 reasons, unfortunately 2 of them didn't really work out either:
* I wanted to avoid having to learn and use yet another domain-specific language to do efficient data wrangling as required in Python. Unfortunately, Polars makes it difficult to use standard Rust tools, so there is still a lot of that here as well.
* We noticed that the Python program became difficult to abort when we ran out of system resources. The hope was that this was due to some inter-process communication or something going on with the Python polars API to core polars (which is implemented in rust). If there was indeed an improvement in this regard, it was not noticeable to me.
* Debugging can be easier in Rust, if you're comfortable with using panics. This is because it embeds the location of each potential panic, and when it happens, displays that location. Polars make debugging difficult due to their lazy and declarative nature. This is the one thing where it worked. Unfortunately, there are still situations where Polars panics internally, which are hard to trace.

Because we're working with a lot of data that may or may not be what we expected, this code is written in a very script-like fashion.

## Installation
I suggest installing rust and its related tools via `rustup`.
Install `rustup` either through your package manager (`homebrew`, `winget`, `apt`, etc. probably have it) or from https://rustup.rs/.

Rustup should install all the tools needed to run the local code on your machine.

If you want to run the code on the cluster, make sure you can compile code that the cluster can run by running `rustup toolchain install stable-x86_64-unknown-linux-gnu` (this is fine everywhere, and required on Apple Silicon or Windows).

The results of both the local and cluster code are mostly already in `./GENERATED/`, so there is no need to install the tools and run the lengthy cluster code.

Update: I wont release the ENEM texts, as they were provided by Marinho and they didnt open them themselves. Theres a [file without this text](./GENERATED/localGen_result_no_secrets.csv) but the identifiers used by Marinho when he sent the data over to us, if you get the same format we got it should be easy to unify.

## Running the local code

The local code is defined in `./src/local.rs' and also uses stuff in `./src/lib.rs'.
Make sure all paths are correct (otherwise you will get a runtime error) and execute:

```sh
cargo run --release --bin local
```
Cargo is Rust's official build system (aka CMake) and package manager (aka npm).

`--release` enables code optimizations.

`--bin local` refers to this section in the [`Cargo.toml`](Cargo.toml)

```toml
[[bin]]
name="local"
path="src/local.rs"
```

On the first run, this will download and compile the dependencies. Due to enabled performance optimizations (mostly for cluster aggregation) this will take a long time, subsequent runs should have them cached.

## Running the Cluster code

Please read the [the local code](#running-the-local-code) section first for more information on what happens.

The bwUniCluster cluster uses x86_64 CPUs and a Linux OS, so we need to compile for that target:
```sh
cargo build --target x86_64-unknown-linux-gnu --release --bin cluster
```

Rust does not dynamically link against many libraries by default, but it does link against gnu C libraries (`libc`, `libgcc_s`, `libm`), as do most binaries.
Getting around this with `musl` was investigated, but caused a lot of other problems.
The bwUniCluster only has old versions of these (from 4 years ago for `libc`) that are no longer supported by Rust, and I could not find an easy way to add/use newer versions of these particular libraries on the cluster.
So I ran the binary in a container that provides modern libraries.
The only container runtime provided on bwUniCluster is `enroot`, so that's what I used.

Alternatively, you could just build the code on the cluster, but it doesn't have Rust tools - so it would need a container anyway - and compiler feedback would be annoying this way, youd have to build on both devices anyway to do debugging.

All steps to execute the bwUniCluster code are in [`execute_cluster.sh`](execute_cluster.sh) (see the [script in model code](../../reimplement/cluster_push_and_queue.sh) for some comments).
(TODO: this is untested as written)

## A few words about Rust
Rust doesnt have exceptions, instead relying on a `Result` type as return value, and `panic`s.
`panic`s are not intended to be cought and handled during program execution, instead they will terminate the program.
`Result` is an enum, but Rust enums differ from eg. Java enums, in that they may contain values per instance (this is know as tagged union, discriminated union, product type).
Since it is a value containing the "good path" value, it behaves similarly to a checked exception in Java, though it is closer to the `Either` class in Haskell - with some additional meaning, the variants are called `Ok` and `Err`.
A `Result` can be panicked with e.g. `unwrap()`, which is often done due to the script-like nature of this code.

Rust also has the `Option` type, which is similar but replaces the concept of (type-checked) nulls in Rust - it is equivalent to `Maybe` in Haskell. This type also has an `unwrap()` which behaves similarly.

One last thing I've done a lot in this code that people may not be aware of is the use of `let-else' bindings.
I will assume familiarity with pattern matching, there are a lot of resources on it, and I've also done it in Python.
Rust has pattern matching, but in many places, e.g. a `let` pattern binding, it requires that the bound pattern is always true, so matching a vector that is known to the programmer but not to the compiler to contain 2 elements is still impossible.

The `let-else` bindings have been adopted in Rust from Swift to allow the use of these `fallible` bindings.
It takes the form: 

```rs
let [element1, element2] = arraylike_object else {
    // some diverging code!
}
```

The `diverging` path must leave the scope into which the pattern is being expanded.
This can be done with a `return', or with a `panic', as this terminates the program and Rust knows about it.
That's why I've mostly put `unreachable!()` in there, which causes a `panic`.

`unreachable!()` and also `dbg!()` are macros.
Rust macros are relatively easier to understand and debug than, say, C macros, and they allow some nice things. For example, `dbg!()` does println debugging on steroids, as it prints not only the inserted value, but also the snippet of code it contains, and its concrete source location.

Rust allows you to assign many things to other things, e.g. if-else, they are all `expressions'.
In particular, blocks of code delimited by `{}` can be placed anywhere (similar to C), but they can also be assigned as expressions.
In this case, and in general in functions, an element that is not followed by a `;` will be returned.
In functions, this is something that is not required. 
But `return` always returns from the current function, so in `{}`-delimited blocks return will not work, so this "implicit return" has to be used.
I use these blocks in some assignments where I want to make sure that the stuff I write in this block is not accidentally picked up elsewhere, but mostly I use these blocks extensively without assigning them.
When I do the latter, chances are that these are tests being run for debugging purposes.
This code is not strictly necessary to understand the data flow, although you may find it useful (I certainly did) to know what the data actually looks like at this point.

For an full intro into Rust I recommend [the excellent and free Rust book](https://doc.rust-lang.org/stable/book/) and/or [rustlings excercises](https://github.com/rust-lang/rustlings) for practice.

## Polars

The only (direct and broad API) dependency of this project is `polars`, a `pandas` alternative written in Rust, with a Python API (used in the related projects).
It has some features that `pandas` lacks, such as support for streaming.
Unfortunately, these specific features are not very mature yet.
In general, the `polars` API does not seem to be very stable at this point, and is also very much oriented towards the Python API, which means that it mostly doesn't take advantage of the niceties that Rusts strong type system can provide, while still requiring strong typing in other situations, which can make some things awkward.

I'm almost certain that I've used a number of APIs in ways that the authors didn't expect, but didn't fail immediately, making it very hard to debug.
The result is some awkward code and some very annoyed comments, sorry about that.

During my time the Rust API already changed.
It was easy to fix these errors, once the replacement was found.
However with a longer timeframe this may get more difficult.