//! This file defines what is run locally.

use std::ops::Deref;

use polars::{lazy::dsl, prelude::{self as pl, LazyFileListReader as _, IntoLazy as _, SerWriter as _, NamedFromOwned, Literal}, series::IntoSeries};
use enem_aggregate::ENEM_QUESTION_ANSWER_OPT;

fn main() {
    let mut _test_csv = pl::CsvWriter::new(std::fs::File::create("./GENERATED/localGen_test.csv").unwrap());

    let question_meta = parse_official_question_meta("../raw_data/enem/official_data");

    // {
    //     let mut laaa = question_meta.clone().lazy().filter(dsl::col("year").eq(dsl::lit(2017))).collect().unwrap();
    //     print!("{laaa}");

    //     _test_csv.finish(&mut laaa).unwrap();
    // }

    let marinho_question_and_answer_text = parse_marinho_enem_data("../raw_data/enem/plaintext_questions_and_irt/enem.csv");

    let interactions_agg = pl::LazyCsvReader::new("./GENERATED/clusterGen_interactions_agg.csv").finish().unwrap();
    
    let mut result_csv = pl::CsvWriter::new(std::fs::File::create("GENERATED/localGen_result.csv").unwrap());
    let mut result_csv_no_secrets = pl::CsvWriter::new(std::fs::File::create("GENERATED/localGen_result_no_secrets.csv").unwrap());

    // These elements together uniquely identify a question (at least in the data we had).
    let question_meta_and_interact = combine_official_question_meta_and_aggregated_interaction_data(question_meta, interactions_agg);

    dbg!(question_meta_and_interact.clone().collect().unwrap());

    let question_meta_and_interact = add_idomatic_ordering(question_meta_and_interact);
    
    let question_and_interact = combine_official_and_aggregated_with_marinho(question_meta_and_interact, marinho_question_and_answer_text);

    dbg!(question_and_interact.clone().collect().unwrap());

    let question_and_interact = question_and_interact.clone()
        .with_column((dsl::col("#correct_answers").cast(pl::DataType::Float64) / dsl::col("#total_answers")).alias("correctness"))
        .select(&[
            "topic",
            "qname",
            "question_and_answer_text",
            "question_id",
            "correctness",
            "^.*irt.*$",
            "anulled",
            "#total_answers",
        ].map(dsl::col))
    ;

    dbg!(question_and_interact.clone().collect().unwrap());

    result_csv.finish(&mut question_and_interact.clone().collect().unwrap()).unwrap();
    result_csv_no_secrets.finish(&mut question_and_interact.clone().select(&[dsl::all().exclude(["^irt_._marinho$", "question_and_answer_text"])]).clone().collect().unwrap()).unwrap();

    // TODO: ENEM and Marinho IRT parameters dont match. At all.
    // which is kind of expected, Marinho describes explicitly that he recovers these parameters
    // And Marinho parameters kindof make sense: they are of a similar order of magnitude as the capability of students
    // (Its hard to identify, but they most likely used either NU_NOTA_REDACAO - writing ability in the writing test before - , OR the resulting grade NU_NOTA_{TOPIC})
    // None of these capabilities explicitly mention IRT in their description (at least the ones I looked at, 2009 + 2016)
    // These parameters should be able to differ without outright conflicting, assuming a different(ly scaled) student ability:
    // Since the difference between ability and difficulty (irt_b) is scaled with the differentiation (irt_a), 
    // if differentiation is scaled down with a factor, and ability and difficulty are scaled up with the same factor, the result is the same
    // But they dont seem to scale like that!
    // the influence of c is also there, and combined with the numerical optimization this might explain some discrepancy
    // Add to this, that the IRT params of the question items are probably just the ones determined on the pretesting, while Marinho regressed on the actual application...

    // FIXME: The thing is, this means I cant really use the difficulty that marinho calculated in my work, can I? I have not created them myself, and I dont have the formula available.
    // best thing I can do is hope that they didnt mess up in terms of correctness, but also, at least the connected bits to this are also just not reproducible, should I use marinho difficulty

    // FIXME: Newer developement: Marinho contains multiple questions, that were "abandoned" (IN_ITEM_ABAN, see excel) from ENEM(INEP) as they didnt allow IRT to converge on their side.
    // these dont have any INEP IRT score attached (but marinho)

    // FIXME: irt_c should always be the same with ENEM, as we've always got 5 choices, so random chance is the same. But NEITHER INEP nor Marinho have them converge even to a roughtly similar size for all questions
    // according to wikipedia that ought to be p(-inf) = c for that same formula, so 1/5 =  0.2
    // Analyzing data:
/*
│ describe   ┆ irt_a    ┆ irt_b    ┆ irt_c    ┆ irt_a_marinho ┆ irt_b_marinho ┆ irt_c_marinho │
│ ---        ┆ ---      ┆ ---      ┆ ---      ┆ ---           ┆ ---           ┆ ---           │
│ str        ┆ f64      ┆ f64      ┆ f64      ┆ f64           ┆ f64           ┆ f64           │
╞════════════╪══════════╪══════════╪══════════╪═══════════════╪═══════════════╪═══════════════╡
│ count      ┆ 1754.0   ┆ 1754.0   ┆ 1754.0   ┆ 1754.0        ┆ 1754.0        ┆ 1754.0        │
│ null_count ┆ 28.0     ┆ 28.0     ┆ 28.0     ┆ 1.0           ┆ 1.0           ┆ 1.0           │
│ mean       ┆ 2.129609 ┆ 1.57606  ┆ 0.163521 ┆ 0.021735      ┆ 653.794005    ┆ 0.141196      │
│ std        ┆ 0.941571 ┆ 1.852909 ┆ 0.068605 ┆ 0.012599      ┆ 117.705627    ┆ 0.080214      │
│ min        ┆ 0.01929  ┆ -2.10568 ┆ 0.00083  ┆ 0.000605      ┆ 325.0         ┆ 0.01          │
│ max        ┆ 8.00279  ┆ 61.05459 ┆ 0.5      ┆ 0.1           ┆ 1010.0        ┆ 0.34          │
│ median     ┆ 2.01504  ┆ 1.46911  ┆ 0.16333  ┆ 0.0192985     ┆ 652.673       ┆ 0.148987      │
│ 25%        ┆ 1.47905  ┆ 0.8782   ┆ 0.12074  ┆ 0.0139272     ┆ 569.096       ┆ 0.0935926     │
│ 75%        ┆ 2.69234  ┆ 2.04605  ┆ 0.20359  ┆ 0.0259177     ┆ 729.578       ┆ 0.195858      │
 */
    // neither Marino nor INEP irt_c look great. Marinho is more contained without quite so large outliers, but mean and median as well as standard deviation are worse

    // FIXME: irt_b Marinho vs INEP from my work doesnt look like his paper AT ALL.
    // INEP irt_b also has LARGE outliers, which I checked back in the original data and were also present in regularly used questionaires (I mean in commonly used CO_PROVA)
    // these questions were also not to my knowledge canceled, and were asked in years (2013 or 2016 i beliefe) where no questions were canceled according to Marinho

    // TODO: Visual inspection of INEP IRT_b vs Marinho IN HIS PAPER has high symmetry of window - and shown data seems symmatric in mean - , suggesting that INEP scores were linearly transformed to same scope
    // Wikipedia suggests looking at ability (and thus irt_b!) as a normal distribution, and plots show 'mean' ability being 0.
    // It might make sense to center like that for numerical stability with floats during regression
    // this would make it good practice, Marinho might not have done that. But with the numerical ranges I've seen Marinho should still be just fine
    // Update: Mean of difficulty for INEP is not 0, its 1.8. Doesnt fit whith them having transformed it, unless that tranformation was calibrated on much less able people(?) beforehand

    // pl::CsvWriter::new(std::fs::File::create("test.csv").unwrap()).finish(&mut res.clone()
    //     .select(&["question_id",/*  "question_ident",*/ "idomatic_ordering_ident"].map(dsl::col))
    //     .explode(&["idomatic_ordering_ident"].map(dsl::col))
    //     .unnest(["idomatic_ordering_ident"])
    //     .sort_by_exprs(&[dsl::col("year"), dsl::col("idomatic_placement_all_topics")], &[false, false], false)
    //     .collect().unwrap()).unwrap();

    // let enem = pl::LazyCsvReader
    //     ::new(format!("../../../raw_data/brasilian/with_corresponding/enem2.csv")).finish().unwrap()
    //     .select([dsl::col("qname")])
    //     .filter(dsl::col("qname").str().starts_with(dsl::lit("ENEM02")).not())
    //     .sort("qname", Default::default())
    // ;

    // dbg!(enem.collect().unwrap());
}

fn combine_official_and_aggregated_with_marinho(question_meta_and_interact: pl::LazyFrame, marinho_data: pl::LazyFrame) -> pl::LazyFrame {
    let question_meta_and_interact = question_meta_and_interact.with_column((
            // determine the identifier as provided by Marinho
            dsl::lit("ENEM")
            + dsl::when(dsl::col("year").eq(dsl::lit(2020))).then(dsl::lit("1")).otherwise(dsl::lit("0"))
            + dsl::col("year").cast(pl::DataType::Utf8)
            + dsl::col("idomatic_placement_all_topics")
                .map(|series| {
                    let iter = series.i64().unwrap().into_iter().map(|el| {
                        let el: i64 = el.unwrap();
                        format!("Q{el:04}")
                    });
                    Ok(Some(pl::Series::from_iter(iter)))
                }, dsl::GetOutput::from_type(pl::DataType::Utf8))
        ).alias("qname"))
        .sort("qname", Default::default());

    dbg!(question_meta_and_interact.clone().collect().unwrap());

    let qname_join = [dsl::col("qname")];
    
    return question_meta_and_interact
        .join(marinho_data, &qname_join, &qname_join, pl::JoinType::Left);
}

fn add_idomatic_ordering(question_meta_and_interact: pl::LazyFrame) -> pl::LazyFrame {
    let idomatic_order_id = {
        // The "idomatic" order is the yellow regular ordering.
        // AFAICT there is no programmatic, somewhat easy to follow and reliable way to find that in the CSVs.
        let mut idomatic_order_id = std::collections::HashMap::new();
        for ((year, topic), co_prova) in [
            ((2009, "CN"), 50),
            ((2009, "CH"), 54),
            ((2009, "LC"), 57),
            ((2009, "MT"), 61),

            ((2010, "CN"), 90),
            ((2010, "CH"), 86),
            ((2010, "LC"), 93),
            ((2010, "MT"), 97),
        
            ((2011, "CN"), 122),
            ((2011, "CH"), 118),
            ((2011, "LC"), 125),
            ((2011, "MT"), 129),

            ((2012, "CN"), 142),
            ((2012, "CH"), 138),
            ((2012, "LC"), 145),
            ((2012, "MT"), 149),

            ((2013, "CN"), 172),
            ((2013, "CH"), 168),
            ((2013, "LC"), 175),
            ((2013, "MT"), 179),

            ((2014, "CN"), 200),
            ((2014, "CH"), 196),
            ((2014, "LC"), 203),
            ((2014, "MT"), 207),

            ((2015, "CN"), 236),
            ((2015, "CH"), 232),
            ((2015, "LC"), 239),
            ((2015, "MT"), 243),

            ((2016, "CN"), 292),
            ((2016, "CH"), 296),
            ((2016, "LC"), 300),
            ((2016, "MT"), 304),

            ((2017, "CN"), 392),
            ((2017, "CH"), 396),
            ((2017, "LC"), 400),
            ((2017, "MT"), 404),

            ((2018, "CN"), 448),
            ((2018, "CH"), 452),
            ((2018, "LC"), 456),
            ((2018, "MT"), 460),

            ((2019, "CN"), 504),
            ((2019, "CH"), 508),
            ((2019, "LC"), 512),
            ((2019, "MT"), 516),

            ((2020, "CN"), 598),
            ((2020, "CH"), 568),
            ((2020, "LC"), 578),
            ((2020, "MT"), 588),

            ((2021, "CN"), 910),
            ((2021, "CH"), 880),
            ((2021, "LC"), 890),
            ((2021, "MT"), 900),
        ] {
            idomatic_order_id.insert((year, topic), co_prova);
        }
        idomatic_order_id
    };

    let question_meta_and_interact = question_meta_and_interact
        // .filter(dsl::col("total_answers").is_not_null())
        .with_column(
            // all this gigant mess does is add a struct row, 
            // that contains the year, and what would be considered the "idomatic" placement of this question 
            // (so the placement in the regular yellow test)
            dsl::col("question_ident").map(move |arg| {
                let arg = arg.list().unwrap().into_iter().map(|list| {
                    let list = list.unwrap();
                    let list = list.struct_().unwrap();
                    let (year, id_pl_all_top): (Vec<_>, Vec<_>) = list.into_iter()
                        .filter_map(|el| {
                            let [
                                ordering_id, 
                                topic, 
                                year, 
                                idomatic_placement_all_topics, 
                                idomatic_placement
                            ] = el else {unreachable!()};
                            // I would be able to combine this pattern with the above, but then I get less fine grained "errors" (unreachable triggering)
                            let pl::AnyValue::Int64(ordering_id) = ordering_id else {unreachable!()};
                            let pl::AnyValue::Int64(year) = year else {unreachable!()};
                            let topic = topic.get_str().unwrap();
                            // I sadly didnt properly document this beforehand,
                            // But I was either very annoyed or sad. (probably both).
                            // I THINK this clause was required due to polars behavior
                            // if one does use fetch with some specific operations, presumable because theye manipulate row numbers (=> the thing fetch "subsamples").
                            // in that case some things end up with nulls.
                            // I did not find this behavior documented anywhere in the official polars documentation.
                            // 
                            // Because I and will want to continue to be able to test this code on stuff generaded by th cluster code with fetch "subsampling"
                            // I had to add this line.
                            //
                            // Take all of this with a pinch of salt!
                            let pl::AnyValue::Int64(mut idomatic_placement_all_topics) = idomatic_placement_all_topics else {
                                // dbg!(el);
                                return None;
                            };
                            let pl::AnyValue::Int64(idomatic_placement) = idomatic_placement else {unreachable!()};
                            // DATA_ERROR: in 2017 they decided NOT to provide the total order in
                            // CO_POSTACIO, instead its the per-topic placement
                            if *year == 2017 {
                                assert_ne!(topic, "LC", "If we handle LC we need some more special casing for that!");
                                let ( _, topic_start_placement) = [("LC", 1), ("CH", 46), ("CN", 91), ("MT", 136)]
                                    .into_iter()
                                    .find(|(pot_topic, _)| *pot_topic == topic).unwrap()
                                ;
                                idomatic_placement_all_topics = topic_start_placement + idomatic_placement;
                            }
                            // dbg!(idomatic_order_co_prova.get(&dbg!((*year, topic))).unwrap());
                            // dbg!(ordering_id);
                            if idomatic_order_id.get(&(*year, topic)).unwrap() != ordering_id {return None};
                            Some((*year, idomatic_placement_all_topics))
                        })
                        .unzip()
                    ;
                    // what do i even name these things for if polars is just going to throw that away like a wet paper towel, and give me ununderstandable errors.
                    pl::StructChunked::new("year_and_idomatic_placement", &[
                        pl::Series::from_vec("year", year), 
                        pl::Series::from_vec("idomatic_placement_all_topics", id_pl_all_top),
                    ]).unwrap().into_series()
                });
                let arg = pl::ListChunked::from_iter(arg).into_series();
                Ok(Some(arg))
            }, Default::default()).alias("year_and_idomatic_placement")
        );

    dbg!(question_meta_and_interact.clone().collect().unwrap());

    let question_meta_and_interact = question_meta_and_interact.clone()
        .filter(dsl::col("year_and_idomatic_placement").list().lengths().neq(dsl::lit(0)))
        .with_column(dsl::col("year_and_idomatic_placement").list().first())
        .drop_columns(&["year"])
        .unnest(["year_and_idomatic_placement"])
        // without this the folowing code wont work.... thanks polars! You only wasted about another 1.5h of my time.
        .collect().unwrap().lazy()
    ;
    question_meta_and_interact
}

fn parse_marinho_enem_data(enem_question_and_answer_text_path: &str) -> pl::LazyFrame {
    let question_bodies = pl::LazyCsvReader
        ::new(enem_question_and_answer_text_path).finish().unwrap()
        .select(&
            [
                dsl::col("qname"),
                dsl::col("anulada").alias("anulled"),
                dsl::col("text").alias("question_and_answer_text"),
            ].into_iter().chain(
                ["a", "b", "c"].map(|irt_param| dsl::col(irt_param).alias(&format!("irt_{irt_param}_marinho")))
            ).collect::<Vec<_>>()
        )
    ;

    dbg!(question_bodies.clone().collect().unwrap());
    question_bodies
}

fn combine_official_question_meta_and_aggregated_interaction_data(
    question_meta: pl::DataFrame, 
    interactions_agg: pl::LazyFrame,
) -> pl::LazyFrame {
    let question_ident = [
        "#question", 
        "topic",
        "ordering_id",
        "year",
    ].map(|elem| dsl::col(elem));

    dbg!(question_meta.clone());

    let question_meta = question_meta.lazy()
        .with_column(dsl::col("year").cast(pl::DataType::Int64))
        .groupby(&question_ident)
        .agg([
            // these should all be unique! I've tested that with past values.
            dsl::col("correct_answer").unique().first(),
            dsl::col("question_id").unique().first(),
            dsl::col("#question_all_topics").unique().first(),
            // TODO: this doesnt work correctly, unlike previous things. I have NO idea why not, dont ask...
            // I have validated that theres only a single unique parameter with manual checks
            // dsl::col("^irt_.$").assert_single_unique(),
            dsl::col("^irt_.$").unique().first(),
            // dsl::col("^irt_.$").unique().list().lengths().map_alias(|old| Ok(format!("{old}_len"))),
        ])
        // .with_column(dsl::col("^irt_.$").list().lengths())
        // .sort("irt_c", Default::default())
        .sort_by_exprs(
            &question_ident,
            [false, false, false, false], 
            true,
        )
        .collect().unwrap()
        .lazy()
    ;

    dbg!(question_meta.clone().collect().unwrap());

    // test_csv.finish(&mut question_meta.clone().collect().unwrap()).unwrap();
    
    // TODO: try reversing
    let question_meta_and_interact = interactions_agg.clone()
        .join(
            question_meta.clone(),
            &question_ident,
            &question_ident,
            pl::JoinType::Left,
        )
        .groupby(&[
            "question_id",
            "topic",
        ].map(dsl::col))
        .agg(&[
            dsl::col("year").unique(),
            dsl::as_struct(&[
                dsl::col("ordering_id"),
                dsl::col("topic"),
                dsl::col("year"),
                dsl::col("#question_all_topics"),
                dsl::col("#question"),
            ]).alias("question_ident"),
            dsl::col("correct_answer").unique().first(),
            dsl::col("^irt_.$").unique().first(),
            dsl::col("^#._answers$").sum(),
            dsl::col("#total_answers").sum(),
        ])
        .with_column(
            // add a column for the "correct" answers
            match_colname(
                dsl::col("correct_answer"), 
                &ENEM_QUESTION_ANSWER_OPT
                    .map(|option| 
                            (option, dsl::col(&format!("#{option}_answers"))
                    )),
                pl::Null{}.lit(),
            ).alias("#correct_answers"),
        )
    ;
    question_meta_and_interact
}

/// This represents an equivalent of the match expression (aka switch-case) for polars dsl.
/// 
/// Without this one would have it being even more verbose and difficulty to read.
fn match_colname(match_on: dsl::Expr, cases: &[(&str, dsl::Expr)], otherwise: dsl::Expr) -> dsl::Expr {
    let mut expr = otherwise;
    for (case, then_expr) in cases.into_iter() {
        expr = dsl::when(match_on.clone().eq(dsl::lit(case.deref()))).then(then_expr.clone()).otherwise(expr);
    };
    expr
}

/// This parses the official metadata about the questions (`/DADOS/ITENS_PROVA_*.csv`).
/// It mostly does validation, and then transformations, first on the year specific data:
/// * renames of rows into clearer names
/// * removes problematic datapoints
/// * adds some datapoints 
///     * such as question year, determined from file location
///     * as well as #question (quesgion number in its respective topic)
/// 
/// Then it aggregates the data of all years together.
fn parse_official_question_meta(root_dir: &str) -> pl::DataFrame {
    // in 2013 they added a special "ledor" edition,
    // that seems to have had different questions in the same positions (CO_POSICAO), 
    // while keeping the same ordering_id (CO_PROVA).
    let forbidden_editions = [
        188,
        187,
        189,
        190,
    ];

    let mut question_meta = Vec::new();

    for year in 2009..=2021 {
        // dbg!(year);
        let provas = pl::LazyCsvReader
            ::new(format!("{root_dir}/{year}/DADOS/ITENS_PROVA_*.csv"))
            // dont ask me why...
            .with_delimiter(if year != 2021 {b';'} else {b','})
            .with_encoding(pl::CsvEncoding::LossyUtf8)
            .finish()
            .unwrap()
            .filter(dsl::col("CO_PROVA").is_in(dsl::lit(pl::Series::from_iter(forbidden_editions))).not())
        ;
        // dbg!(provas.clone().collect().unwrap());

        let questions = provas.clone()
            .select(&[
                dsl::col("CO_ITEM").alias("question_id"),
                dsl::col("TX_GABARITO").alias("correct_answer"),
                dsl::col("SG_AREA").alias("topic"),
                dsl::col("NU_PARAM_A").alias("irt_a"), // aka differential (not named this way for consistency with data from other sources)
                dsl::col("NU_PARAM_B").alias("irt_b"), // aka difficulty
                dsl::col("NU_PARAM_C").alias("irt_c"), // aka guess
                dsl::col("TX_COR").alias("color"),
                dsl::col("CO_POSICAO").alias("#question_all_topics"),
                dsl::col("CO_PROVA").alias("ordering_id"),
            ])
            .with_columns([
                dsl::col("color").str().to_lowercase(),
                dsl::lit(year).alias("year"),
            ])
            // filter out weirdness of language sectionin said topic: 
            // It may have questions with the same index but different languages,
            // also the different language may be confusing to non-multilanguage LMs
            .filter(dsl::col("topic").neq(dsl::lit("LC")))
        ;

        {
            let unique_per_question = ["correct_answer", "topic", "irt_a", "irt_b", "irt_c"];

            let questions = questions.clone()
                .groupby([dsl::col("question_id")])
                .agg(unique_per_question.map(|col| dsl::col(col).unique()))
                .collect().unwrap()
            ;
            for unique_field in unique_per_question {
                for (idx, entry) in questions.clone().column(unique_field).unwrap().iter().enumerate() {
                    let pl::AnyValue::List(entry) = entry else {unreachable!()};
                    if entry.len() != 1 {
                        dbg!(questions.clone().get(idx));
                        // DATA_ERROR: Spotted an error?
                        // CO_ITEM = 58542 has different answer SPECIFICALLY for TX_COR = ROSA 
                        assert_eq!(entry.len(), 1);
                    }
                }
            }
        }

        // dbg!(questions.clone().collect().unwrap());

        let questions = questions
            .groupby(&[
                dsl::as_struct(&[
                    dsl::col("year"),
                    dsl::col("topic"), // duplicated with ordering_id, but this is helpful for processing the interactions
                    dsl::col("#question_all_topics"),
                    dsl::col("ordering_id"),
                ]).alias("occurence_data"),
            ])
        ;


        {
            let unique_per_occurence = ["color", "question_id"];

            let questions_1 = questions.clone()
                .agg(unique_per_occurence.map(|col| dsl::col(col).unique()))
            ;

            let questions = questions_1.clone()
                .collect().unwrap()
            ;

            // dbg!(questions_1.clone().filter(dsl::col("question_id").list().lengths().neq(dsl::lit(1))).collect().unwrap());
            // DATA_ISSUE: In 2013 they break another rule.
            // They did a special "Ledor" edition of a random color per topic,
            // where they have different questions in the same spots for the same ordering_id.
            // This is now fixed with filtering out these questions.

            for unique_field in unique_per_occurence {
                for (idx, entry) in questions.clone().column(unique_field).unwrap().iter().enumerate() {
                    let pl::AnyValue::List(entry) = entry else {unreachable!()};
                    if entry.len() != 1 {
                        dbg!("Repetition of {unique_field}!");
                        dbg!(questions.clone().get(idx));
                        assert_eq!(entry.len(), 1);
                    }
                }
            }
        }

        let questions = questions.clone()
            .agg(&[
                dsl::as_struct(&[
                    dsl::col("question_id"),
                    dsl::col("correct_answer"),
                    dsl::col("^irt_.*$"),
                ]).alias("question_meta"),
            ])
        ;

        {
            let questions = questions.clone().collect().unwrap();
            for (idx, entry) in questions.clone().column("question_meta").unwrap().iter().enumerate() {
                let pl::AnyValue::List(entry) = entry else {unreachable!()};
                if entry.len() != 1 {
                    dbg!("Repetition of combined question identifier!");
                    dbg!(questions.clone().get(idx));
                    assert_eq!(entry.len(), 1);
                }
            }
        }

        let questions = questions.with_column(dsl::col("question_meta").list().first());

        // dbg!(questions.clone()
        //     .unnest([
        //         "occurence_data", 
        //         "question_meta",
        //     ])
        //     .collect().unwrap()
        // );

        let questions_by_ordering = questions.clone()
            .unnest([
                "occurence_data", 
                // "question_meta",
            ])
            .sort_by_exprs(
                [dsl::col("#question_all_topics")],
                [false], false
            )
            .groupby([
                dsl::col("ordering_id")
            ])
            .agg([
                dsl::as_struct(&[
                    dsl::col("ordering_id"),
                    (
                        dsl::col("#question_all_topics") - dsl::col("#question_all_topics").min()
                    ).alias("#question"),
                    dsl::col("topic"),
                    dsl::col("year"),
                    dsl::col("#question_all_topics"),
                ]).alias("occurence_data"),
                dsl::col("question_meta"),
            ])
            // already present in occurence_data
            .drop_columns(["ordering_id"])
        ;

        question_meta.push(questions_by_ordering.clone());

        // dbg!(questions_by_ordering.clone().collect().unwrap());
    }

    let question_meta = pl
        ::diag_concat_lf(
            question_meta,
            true,
            false,
        )
        .unwrap()
        .explode([
            dsl::col("occurence_data"),
            dsl::col("question_meta"),
        ])
        // FIXME: for some reason this + select panics...
        // .unnest([
        //     "occurence_data",
        //     "question_meta",
        // ])
        // .select(&[
        //     dsl::col("topic"),
        //     dsl::col("#question"),
        //     dsl::col("ordering_id"),
        //     dsl::col("year"),
        //     dsl::col("correct_answer"),
        //     dsl::col("question_id"),
        //     dsl::col("#question_all_topics"),
        //     dsl::col("^irt_.$"),
        // ])
        .select(&[
            dsl::col("occurence_data").struct_().field_by_name("topic"),
            dsl::col("occurence_data").struct_().field_by_name("#question"),
            dsl::col("occurence_data").struct_().field_by_name("ordering_id"),
            dsl::col("occurence_data").struct_().field_by_name("year"),
            dsl::col("occurence_data").struct_().field_by_name("#question_all_topics"),
            // FIXME: Hotfix for a really weird error when accessing fields 1. by name in any way 2. by any number other than 0
            dsl::col("question_meta").struct_().rename_fields(vec!["question_id".into(), "correct_answer".into(), "irt_a".into(), "irt_b".into(), "irt_c".into()]).struct_().field_by_name("correct_answer"),
            dsl::col("question_meta").struct_().rename_fields(vec!["question_id".into(), "correct_answer".into(), "irt_a".into(), "irt_b".into(), "irt_c".into()]).struct_().field_by_name("question_id"),
            dsl::col("question_meta").struct_().rename_fields(vec!["question_id".into(), "correct_answer".into(), "irt_a".into(), "irt_b".into(), "irt_c".into()]).struct_().field_by_name("irt_a"),
            dsl::col("question_meta").struct_().rename_fields(vec!["question_id".into(), "correct_answer".into(), "irt_a".into(), "irt_b".into(), "irt_c".into()]).struct_().field_by_name("irt_b"),
            dsl::col("question_meta").struct_().rename_fields(vec!["question_id".into(), "correct_answer".into(), "irt_a".into(), "irt_b".into(), "irt_c".into()]).struct_().field_by_name("irt_c"),
        ])
        .collect().unwrap()
    ;

    dbg!(question_meta.clone());

    {
        // more assertions to catch gotchas
        // DATA_ERROR: Glad I did that. It seems question_id also isnt really unique. Theres 1 case
        // question_id/CO_ITEM = 60227
        // appears in 2009 in MT and 2017 in CN
        let questions_by_q = question_meta.clone().lazy()
            .groupby([dsl::col("question_id")])
            .agg([
                dsl::col("topic").unique(),
                dsl::col("#question"),
                dsl::col("year").unique(),
                dsl::col("correct_answer").unique(),
                dsl::col("ordering_id"),
            ])
            .collect().unwrap()
        ;

        let unique = [
            "topic", 
            // "year", 
            "correct_answer",
        ];

        for unique_field in unique {
            for (idx, entry) in questions_by_q.clone().column(unique_field).unwrap().iter().enumerate() {
                let pl::AnyValue::List(entry) = entry else {unreachable!()};
                if entry.len() != 1 {
                    // Do consider that every repetition will result in 2 errors ;-)
                    dbg!("Repetition of CO_ITEM!");
                    dbg!(idx);
                    dbg!(questions_by_q.clone().get(idx));
                }
            }
        }
    }
    question_meta
}
