"""
Contains functions to actually parse moodles xml question format (using `moodle_questions_dataclasses.py`).

If executed as main it tests these methods, and also determines some statistics.
"""
from xml.etree import ElementTree
from common_py.utils import UNREACHABLE
from typing import Iterable, Tuple
from itertools import product
import moodle_questions_dataclasses as mqd

class CollectDebugInfos():
    "Just a little record keeping for debugging purposes"
    questionTypesMetaInfo: dict

    def __init__(self):
        self.questionTypesMetaInfo = {}
    
    def collectData(self, question: ElementTree.Element):
        questionType = question.attrib["type"]
        if questionType not in self.questionTypesMetaInfo:
            self.questionTypesMetaInfo[questionType] = {
                "child_attrib": {},
            }
        debug_meta_info = self.questionTypesMetaInfo[questionType]
        for item in question.findall("./"):
            if item.tag not in debug_meta_info["child_attrib"]:
                debug_meta_info["child_attrib"][item.tag] = {}

def matchAndCreateQuestion(question: ElementTree.Element, schoolClassAsked: str | None):
    common = mqd.CommonQuestionFields.createFromXml(question, schoolClassAsked)
    match question.attrib["type"]:
        case "multichoice":
            return mqd.MultiChoice.createFromXml(question, common)
        case "truefalse":
            return mqd.TrueFalse.createFromXml(question, common)
        case "matching":
            return mqd.Matching.createFromXml(question, common)
        case "ddimageortext":
            return mqd.NotHandled.createFromXml(question, common=common, reason="Very likely to contain/require images")
        # These questions are only present in PGM1+2
        case "cloze":
            return mqd.NotHandled.createFromXml(question, common=common, reason="not a closed question")
        case "gapselect":
            return mqd.Gapselect.createFromXml(question, common)
        # "Move item to the right spot in the sentence", the right choice is encoded in the marker, [[i]] means the ith answer (beginning with 1) is the correct choice.
        case "ddwtos":
            return mqd.Ddwtos.createFromXml(question, common)
        case _:
            UNREACHABLE("Handle all possible question types")

def extractQuestionsFromMoodleQuizXml(
    subjectAndXmlPaths: Iterable[Tuple[str, str]]
) -> mqd.QuestionsDataType:
    ":param subjectAndXmlPaths: Iterator of Tuple Pairs of (subject, xml_path)"
    questionsData: mqd.QuestionsDataType = []

    debugInfo = CollectDebugInfos()

    for (subject, xmlPath) in subjectAndXmlPaths:
        for question in ElementTree.parse(xmlPath).getroot().iter("question"):
            debugInfo.collectData(question)
            questionType = question.attrib["type"]

            # "category" questions are not actually questions it appears
            if not question.find("name"):
                assert questionType == "category"
                continue
            # adding a bunch of assertions to avoid silent failure (also in following match)
            assert question.find("./questiontext").attrib["format"] == "html"
            questionsData.append(matchAndCreateQuestion(question, schoolClassAsked=subject))
    return questionsData

if __name__ == "__main__":
    subjects = ["ASV", "KI", "MLDM", "PGM_1_2"]
    questionsData = extractQuestionsFromMoodleQuizXml(
        map(lambda subject: (subject, f"../raw_data/hft/{subject}/quiz.moodle.xml"), subjects)
    )

    # pprint(questionsData)

    questionTypes = [
        "multichoice",
        "truefalse",
        "matching",
        "ddimageortext",
        "cloze",
        "gapselect",
        "ddwtos",
    ]

    print("Table")
    for (maybeSubject, maybeType) in product([*subjects, None], [*questionTypes, None]):
        print(f"\t{maybeSubject}, {maybeType}: {len([itm for itm in questionsData if (not maybeSubject or itm.common.schoolClassAsked == maybeSubject) and (not maybeType or itm.common.questionType == maybeType)])}\n")

    questionsWithoutFiles = [itm for itm in questionsData if not itm.common.questionContainsFile]
    print("Table")
    for (maybeSubject, maybeType) in product([*subjects, None], [*questionTypes, None]): #"singlechoice", 
        print(f"\t{maybeSubject}, {maybeType}: {len([itm for itm in questionsWithoutFiles if (not maybeSubject or itm.common.schoolClassAsked == maybeSubject) and (not maybeType or (itm.common.questionType == maybeType))])}\n") #and (not maybeType.endswith('choice') or (itm.singleChoice if maybeType == 'singlechoice' else not itm.singleChoice))
        if maybeType and maybeType.endswith("choice"):
            print(f"single: {len([itm for itm in questionsWithoutFiles if (not maybeSubject or itm.common.schoolClassAsked == maybeSubject) and (not maybeType or (itm.common.questionType == maybeType and itm.singleChoice))])}")
            print(f"multi: {len([itm for itm in questionsWithoutFiles if (not maybeSubject or itm.common.schoolClassAsked == maybeSubject) and (not maybeType or (itm.common.questionType == maybeType and not itm.singleChoice))])}")
    # countAll = 0
    # countTrueFalse = 0
    # for question in questionsData:
    #     countAll += 1
    #     if isinstance(question, TrueFalse):
    #         countTrueFalse += 1

    # print(countAll, countTrueFalse)

    for questionData in questionsData:
        match questionData:
            case mqd.Matching(common=mqd.CommonQuestionFields(questionText = questionText)):
                pass
                # UNREACHABLE("testing matching")
