"""
This contains only dataclasses describing the questiontypes 
in moodles xml export -- the ones that were found in my data, theres probably more.

Each dataclass also contains a function that can create itself 
when given a `xml.etree.ElementTree.Element` that corresponds to itself.
"""

from dataclasses import dataclass
import dataclasses
from xml.etree import ElementTree
from math import nan

@dataclass
class CommonQuestionFields():
    name: str
    questionText: str
    "This text is enriched with HTML"
    questionContainsFile: bool
    defaultgrade: float
    penalty: float
    questionType: str
    schoolClassAsked: str | None = None

    @staticmethod
    def createFromXml(question: ElementTree.Element, schoolClassAsked: str | None = None):
        questionType = question.attrib["type"]
        # a way around special casing for close without making other things annoying. This WOULD blow in our face if we were to use it, thankkfully, but we dont want to.
        try:
            defaultgrade = float(question.find("./defaultgrade").text)
        except:
            assert questionType == "cloze"
            defaultgrade = nan
        return CommonQuestionFields(
            name = question.find("./name/text").text,
            questionText = question.find("./questiontext/text").text,
            questionContainsFile = question.find("./questiontext/file") is not None,
            defaultgrade = defaultgrade,
            penalty = float(question.find("./penalty").text),
            schoolClassAsked = schoolClassAsked,
            questionType=questionType,
        )

@dataclass
class MultiChoice():
    @dataclass
    class Answer():
        correctFract: float | None
        "None is the answer is wrong, float if the answer is correct and contributes its value of the total points"
        text: str

    singleChoice: bool
    common: CommonQuestionFields
    answers: list[Answer]

    @staticmethod
    def createFromXml(question: ElementTree.Element, common: CommonQuestionFields):
        # AFAICT multichoice questions mark correct answers with a fraction != 0, and theres no other marker
        answers = question.findall("./answer")
        for ans in answers:
            assert ans.attrib["format"] == "html"
            assert ans.find("file") is None
        answers = [
            (float(ans.attrib["fraction"]) / 100, ans.find("text").text)
            for ans in answers
        ]
        return MultiChoice(
            common=common, 
            # found 2 questions where there was only one correct answer but it was not marked as single question
            # Originally these were also considered single choice questions, and added with
            #  or sum([frac == 1 for (frac, _) in answers]) == 1
            # However, on reconsideration this was not correct.
            # The KNOWLEDGE that a question only has a single true answer can strongly influence performance.
            singleChoice=question.find("./single").text == "true",
            answers= [
                MultiChoice.Answer(
                    correctFract= float(ans[0]) if (int(ans[0]) > 0) else None,
                    text=str(ans[1]),
                )
                for ans in answers
            ]
        )


@dataclass
class TrueFalse():
    @dataclass
    class Answer():
        correct: bool
        value: bool

    common: CommonQuestionFields
    answers: list[Answer]

    @staticmethod
    def createFromXml(question: ElementTree.Element, common: CommonQuestionFields):
        answers = question.findall("./answer")
        for ans in answers:
            assert ans.attrib["format"] == "moodle_auto_format"
            assert ans.find("file") is None
        answers = [
            (float(ans.attrib["fraction"]) / 100, ans.find("text").text)
            for ans in answers
        ]
        return TrueFalse(
            common=common, 
            answers= [
                TrueFalse.Answer(
                    correct = fract > 0.5,
                    value = text == "true",                            
                )
                for (fract, text) in answers
            ]
        )


@dataclass
class Matching():
    @dataclass
    class SubQuestion():
        text: str
        answer: str
    
    common: CommonQuestionFields
    subQuestions: list[SubQuestion]

    @staticmethod
    def createFromXml(question: ElementTree.Element, common: CommonQuestionFields):
        subquestions = question.findall("./subquestion")
        for subq in subquestions: 
            assert subq.attrib["format"] == "html"
            assert subq.find("file") is None
        return Matching(
            common=common,
            subQuestions=[
                Matching.SubQuestion(text = subq.find("./text").text, answer = subq.find("./answer/text").text) 
                for subq in subquestions
            ],
        )

@dataclass
class Ddwtos():
    """
    `common.questionsText` contains numbered fields `[[i]]`,
    where i indicates the item to insert in that spot
    """
    @dataclass
    class Dragbox():
        text: str
        position: int
    
    common: CommonQuestionFields
    dragboxes: list[Dragbox]

    @staticmethod
    def createFromXml(question: ElementTree.Element, common: CommonQuestionFields):
        dragboxes = question.findall("./dragbox")
        for item in dragboxes: 
            assert item.find("file") is None
        return Ddwtos(
            common=common,
            dragboxes=[
                Ddwtos.Dragbox(
                    text = item.find("./text").text,
                    position = idx,
                )
                for (idx, item) in enumerate(dragboxes)
            ],
        )

@dataclass
class Gapselect():
    """
    `common.questionsText` contains numbered fields `[[i]]`,
    where i indicates the item to select for that spot
    """
    @dataclass
    class SelectOption():
        text: str
        position: int
    
    common: CommonQuestionFields
    selectOptions: list[SelectOption]

    @staticmethod
    def createFromXml(question: ElementTree.Element, common: CommonQuestionFields):
        selectOption = question.findall("./selectoption")
        for item in selectOption: 
            assert item.find("file") is None
        return Gapselect(
            common=common,
            selectOptions=[
                Gapselect.SelectOption(
                    text = item.find("./text").text,
                    position = idx,
                )
                for (idx, item) in enumerate(selectOption)
            ],
        )

@dataclass
class NotHandled():
    name: str
    reason: str
    xml: ElementTree.Element
    common: CommonQuestionFields | None

    @staticmethod
    def createFromXml(question: ElementTree.Element, reason: str, common: CommonQuestionFields | None = None):
        return NotHandled(
            name = question.find("./name/text").text, 
            xml = question,
            reason = reason,
            common = common,
        )

QuestionsDataType = list[MultiChoice | Matching | TrueFalse | Matching | Gapselect | Ddwtos | NotHandled]
