"""
Contains functions to convert the multliple variants of moodle question types to a more generic one.
This makes a few choices in how these are mapped!
"""
from dataclasses import dataclass
import moodle_questions_dataclasses as mdt
from itertools import product
from common_py.utils import UNREACHABLE
from pprint import pprint
from typing import TypedDict

# polars doesnt like lists of dataclasses as elements, 
# but for earlier processing TypedDict is uncomfortable,
# as one cant do type-based match on that
# So I end up using both, oh well.
class AnswerOption(TypedDict):
    text: str
    "Enriched with HTML"
    # recievedPointsFract: float
    correct: bool

@dataclass
class GenericQuestion():
    name: str
    "This is the name from the moodle XML export, not the csv"
    questionText: str
    "Enriched with HTML"
    answerOptions: list[AnswerOption]
    questionType: str
    "The original question type (to allow filtering in case some types are suspected to be problematic)"

def mapToGenericQuestionFormat(
    input: mdt.QuestionsDataType,
):
    trueMultichoiceOptions = []
    matchingOptions = []
    skippedBecauseImage = []
    skippedForAnotherReason = []
    output: list[GenericQuestion] = []
    for elem in input:
        meta = elem.common
        match elem:
            case mdt.NotHandled(name, _):
                skippedForAnotherReason.append(name)
            # Skip questions which contain an image (assumes that all types contain meta as first field)
            case elem if elem.common.questionContainsFile:
                skippedBecauseImage +=[meta.name]
                continue
            case mdt.MultiChoice(answers = answersSrc, singleChoice = singleChoice):
                answers: list[AnswerOption] = []
                if singleChoice:
                    answers.extend([
                        AnswerOption(
                            text = ans.text,
                            # recievedPointsFract=0.0 if ans.correctFract is None else 1.0,
                            correct=False if ans.correctFract is None else True,
                        ) for ans in answersSrc
                    ])
                else:
                    trueMultichoiceOptions += [len(answersSrc)]
                    selections = answersSrc
                    answers.extend([
                        AnswerOption(
                            text = ans.text,
                            correct = False if ans.correctFract is None else True,
                            # recievedPointsFract=0,
                        ) for ans in selections
                    ])
                output.append(GenericQuestion(
                    name=meta.name,
                    questionText=meta.questionText,
                    answerOptions=answers,
                    questionType= "singlechoice" if singleChoice else meta.questionType,
                ))
            case mdt.TrueFalse(answers = answersSrc):
                answers = [
                    AnswerOption(
                        text = "Die Aussage stimmt" if ans.correct else "Die Aussage stimmt nicht",
                        # recievedPointsFract=1.0 if ans.correct else 0.0,
                        correct = ans.correct,
                    ) for ans in answersSrc
                ]
                output.append(GenericQuestion(
                    name=meta.name,
                    questionText=meta.questionText,
                    answerOptions=answers,
                    questionType=meta.questionType,
                ))
            case mdt.Matching(subQuestions = subQuestionsSrc):
                answers: list[AnswerOption] = []
                for ((subqIdx, subquestion), (ansIdx, answer)) in product(
                    enumerate([ans.text for ans in subQuestionsSrc]), 
                    enumerate([ans.answer for ans in subQuestionsSrc])
                ):
                    answers.append(AnswerOption(
                        text = f"{answer} passt zu {subquestion}",
                        correct = subqIdx == ansIdx,
                    ))
                # for matchedIndexes in permutations(range(len(subQuestionsSrc))):
                #     answerTextParts = [
                #         f"{subQuestionsSrc[answerIdx].answer} passt zu {subQuestionsSrc[textIdx].text}" 
                #         for (textIdx, answerIdx) in enumerate(matchedIndexes)
                #     ]
                #     answerPoints = meta.defaultgrade - sum([
                #         textIdx != answerIdx
                #         for (textIdx, answerIdx) in enumerate(matchedIndexes)
                #     ]) * meta.penalty
                #     answers.append(AnswerOption(
                #         text = reduce(lambda rhs, lhs: f"{rhs}, {lhs}", answerTextParts[1:], answerTextParts[0]),
                #         recievedPointsFract=float(answerPoints) / meta.defaultgrade,
                #     ))
                matchingOptions += [len(answers)]
                output.append(GenericQuestion(
                    name = meta.name,
                    questionText=meta.questionText,
                    answerOptions= answers,
                    questionType=meta.questionType,
                ))
            case mdt.Ddwtos(dragboxes=dragBoxesSrc, common=mdt.CommonQuestionFields(name, _)):
                skippedForAnotherReason.append(name)
                # I wont put in that work as this is unlikely to work,
                # considering that the generated text will be too long
                continue
            case mdt.Gapselect(selectOptions = selectOptionsSrc, common=mdt.CommonQuestionFields(name, _)):
                skippedForAnotherReason.append(name)
                # I wont put in that work as this is unlikely to work,
                # considering that the generated text will be too long
                continue
            case _:
                UNREACHABLE("Please be exhaustive (explicitly skip ignored stuff)")
    print(f"""
    #options true multichoice" {trueMultichoiceOptions}
    #options match (also true multichoice) {matchingOptions}
    #skipped because of pictures {len(skippedBecauseImage)}
    #skipped because of other reason {len(skippedForAnotherReason)}
    """)
    return output

if __name__ == "__main__":
    from extract_moodle_xml import extractQuestionsFromMoodleQuizXml

    pprint(
    len(mapToGenericQuestionFormat(extractQuestionsFromMoodleQuizXml(
        map(lambda subject: (subject, f"../raw_data/hft/{subject}/quiz.moodle.xml"), ["ASV", "KI", "MLDM", "PGM_1_2"])
    )))
    )

# FIXME: Determine typical context length for BERT etc, maybe some of the answer serializations are not possible as they result in too much text
# Alternative would be some other encoding, but its doubtful whether the LLM can learn that encoding with out dataset.
# EG keeping the [[1]] fields in [[2]] the sentence and providing an order for answers
# Is there some sparse encoding example in typical pretraining data/tasks?

# TrueFalse: Prepend with "Ist der foldende Satz wahr oder falsch?", use these as variants?
# Matching: subq.text " passt zu " subq.answer
# DDWTOS: "Welcher der folgenden sätze ist korrekt?", options with text filled in

# FIXME: Kein true multiple choice in ENEM (Regex /#T((.*?)\n #){1,8}T/, getested an einem gefaktem positive)
# FIXME: Enthält als einziges itemize|array|footnotesize umgebungen (keine tables)