from typing import Iterable, TypedDict
from transformers import (
    AutoTokenizer,
    TFAutoModel,
    PreTrainedTokenizerBase,
)
from transformers.tokenization_utils import (
    TruncationStrategy,
    PaddingStrategy,
)
import config
import numpy as np
import tensorflow as tf
import os

def create_tf_model(
    huggingface_model: str,
    max_length: int,
    learning_rate: float,
    dropout_regress: float,
    dropout_base: float,
    freeze_base: bool = False,
):
    """
    Composes the base model from huggingface with a single neuron top layer using the [CLS] token embedding to do regression.
    :param huggingface_model: the name of the huggingface identifier
    :param max_length: max sequence length (usually max token lenght of huggingface model)
    :param learning_rate: Adam learning rate used by tensorflow.
    :param dropout_regress: the dropout of the fully connected layer on the top of the transformer [CLS] token embedding. 
    Can be used to fight overfitting.
    :param dropout_base: the dropout for base model
    :param freeze_base: Freeze base model parameters (may be unable to learn) 
    or keep learning (may be destructive)
    """
    # import pre trained model to fine tune
    base_model = TFAutoModel.from_pretrained(huggingface_model, from_pt = True if huggingface_model=='neuralmind/bert-large-portuguese-cased' else False,)
    # we use separately configurable dropout of base model and regression layer
    base_model.config.dropout = dropout_base
    base_model.trainable = not freeze_base
    # gotta have an input for the tf functional API
    input_ids = tf.keras.layers.Input(shape=(max_length,), name='input', dtype=tf.int32)
    # last layer = embedding layer
    embedding_layer = base_model([input_ids])[0]
    # take the only embeddings corresponding to the CLS token which was trained in NSP
    cls_corresp = embedding_layer[:, 0, :]
    # add dropout for final layer
    regression_dropout = tf.keras.layers.Dropout(dropout_regress, name="regression_dropout")(cls_corresp)
    # output layer 1 neuron linear activation function (default one) since we are performing regression
    regression = tf.keras.layers.Dense(1, name="regression")(regression_dropout)
    model = tf.keras.Model(inputs=[input_ids], outputs=regression)
    # compile the model
    model.compile(
        optimizer=tf.optimizers.Adam(learning_rate=learning_rate),
        loss=tf.keras.losses.MeanSquaredError(),
        metrics=[
            tf.keras.metrics.MeanAbsoluteError(),
            tf.keras.metrics.RootMeanSquaredError(),
        ],
    )
    return model

def encode_using(
    texts: Iterable[str],
    huggingface_model: str,
    max_length: int,
):
    tokenizer: PreTrainedTokenizerBase = AutoTokenizer.from_pretrained(huggingface_model)
    return tokenizer(
        list(texts),
        add_special_tokens=True,
        return_token_type_ids=True,
        return_attention_mask=True,
        padding=PaddingStrategy.MAX_LENGTH,  # pads to the right by default
        truncation=TruncationStrategy.LONGEST_FIRST,
    )["input_ids"]

def SIDEEFFECTS_set_seeds(
    seed: int,
):
    """
    I'd really like to avoid this, 
    but ensuring this isnt required would require a round each to test for each global seed set,
    and even in that case it the platform might just have choosen the same seeds for another reason
    that still isnt really stable.

    So rather we stick with this.
    """
    # THIS IS UNTESTED! Didnt get around to testing this, too much other things to do at this point.
    tf.keras.utils.set_random_seed(seed)
    tf.config.experimental.enable_op_determinism()
    # the lines below were taken from the example code from beneditto.
    # Turns out theyre hot garbage.
    # Not only do they not work,
    # they outright setup a failure case when other things come together, 
    # which is kindof nice I guess as this made me look up the proper way.
    # The failure cause is the setting of `TF_DETERMINISTIC_OPS` to a stringified seed.
    # This env variable is evaluated to a boolean, so anything but 0 and 1 as values lead to an abort.
    # The reason this didnt fail for beneditto,
    # is because IT NEVER WAS READ.
    # I triggered a read with my addition of `tf.keras.backend.clear_session()` 
    # to actually clear the memory after repeated model creations.
    # After this was done, tensorflow ACTUALLY read this env variable, and it caused an abort.
    # Thankfully I was aware of all this being black magic 
    # and thus I remembered it vividly when i saw that env variable name in the log.
    # 
    # Had I blindly copy pasted this, I wouldve been on the hunt for a long time.
    # np.random.seed(seed)
    # tf.random.set_seed(seed)
    # os.environ['TF_DETERMINISTIC_OPS'] = str(seed)

def reset_tf_memory():
    """
    thanks tensorflow for being so... interesting(ironically).
    """
    tf.keras.backend.clear_session()
    # the below was also suggested, but doesnt seem to make a terrible much of sense, see:
    # https://www.tensorflow.org/guide/gpu#limiting_gpu_memory_growth
    # I don't share that GPU with anything or anyone, so deallocating memory from the underlying tensorflow allocator, 
    # or however tensorflow manages that stuff, isnt required.
    # Also that link warns about potential memory fragmentation.
    # Fragmentation should PROBABLY not happen, if clear_session doesnt leave stuff behind, 
    # but why bet if I dont have to.
    # tf.config.experimental.set_memory_growth(tf.config.experimental.list_physical_devices('GPU')[0], True)

def SIDEEFFECTS_create_model_with_seed(
    config: config.AllConfig
):
    """
    Ensure to create a new model before each other config run!
    """
    # Since this was wrong during pretty much my entire work,
    # I'm taking no risks here any longer
    SIDEEFFECTS_set_seeds(config.seed)
    reset_tf_memory()
    SIDEEFFECTS_set_seeds(config.seed)

    model = create_tf_model(
        config.model.huggingface_name,
        config.model.context_length, 
        learning_rate = config.training.learning_rate,
        dropout_base = config.training.dropout_pretrained,
        dropout_regress = config.training.dropout_regression,
        freeze_base = config.training.freeze_base,
    )

    model.summary()

    return model

