import numpy as np
import polars as pl
from typing import NamedTuple
from numpy import typing as npt
from common_py.plotting import *
from common_py.utils import Statistics, identity
import re
import os
import json

def get_symmetric_stddev_bins(statistics: Statistics, binsize_multi: float = 1.0):
    """
    Returns a list of bin limits, limits are 'symmetric' around the mean, with length of binsize_multi * std_dev
    """
    # end_mean_bin = mean + std_dev/2.0
    (min, max, mean, std_dev) = statistics
    bin_size = std_dev * binsize_multi
    start_first_bin = np.floor(min / bin_size) * bin_size - bin_size / 2.0
    num_bins = np.ceil((max - start_first_bin) / bin_size)
    return np.arange(num_bins + 1, dtype=np.float64) * bin_size + start_first_bin

if __name__ == '__main__':
    analyzedDir = "./results/containerbasedv5_d_regr"
    # analyzedDir = "./results/enem/eval_paramv3"

    # per_category_train = PlotData(
    #     title='Training progress, comparison for training data',
    #     save_path=f"{basepath}/progress_comparison_train.pdf",
    #     x_ax=AxData('Epoch'),
    #     y_ax=AxData(f'Loss: MSE, logarithmic', scale="log"),
    #     # we want to add that data piece by piece, 
    #     # however with a default value for plot, 
    #     # the different PlotData instances would share a single list.
    #     graphs=[],
    # )
    # per_category_validate = PlotData(
    #     title='Training progress, comparison for validation data',
    #     save_path=f"{basepath}/progress_comparison_validate.pdf",
    #     x_ax=AxData('Epoch'),
    #     y_ax=AxData(f'Loss: MSE, logarithmic', scale="log"),
    #     graphs=[], 
    # )
    for modelConfigIdentifier in [
        match.group(1) 
        for match
        in [
            re.match(r"(hash\[[\d\w-]*\])_config.json", fileName)
            for fileName
            in os.listdir(analyzedDir)
        ]
        if match
    ]:  
        try:
            baseStr = f"{analyzedDir}/{modelConfigIdentifier}"
            with open(f"{baseStr}_config.json") as fp:
                config = json.load(fp)
            model = config["model"]["huggingface_name"]
            history = pl.read_csv(f"{baseStr}_history.csv")
            train_loss = history['loss']
            val_loss = history['val_loss']
            target = config["target"]
            PlotData(
                graphs=[
                    Plot(yData= train_loss, label='Training MSE'),
                    Plot(yData= val_loss, label='Validation MSE'),
                ],
                x_ax=AxData('Epoch'),
                y_ax=AxData('Loss: MSE, logarithmic axis', scale="log"),
                save_path=f"{baseStr}_history.pdf",
            ).create().draw_and_save()
            # per_category_train.graphs.append(Plot(yData=train_loss, label=model))
            # per_category_validate.graphs.append(Plot(yData=val_loss, label=model))

            predictions: pl.DataFrame = (pl.read_csv(f"{baseStr}_predictions.csv", dtypes={"split": pl.Categorical})
                .with_columns(
                (pl.col(target) - pl.col(f"predicted_{target}"))
                    .abs().alias(f"{target}_err_abs"))
            )
            predictions_valid: pl.DataFrame = (predictions
                .filter(pl.col("split") == "validate")
            )
            predictions_train: pl.DataFrame = (predictions
                .filter(pl.col("split") == "train")
            )
            pred_descr: pl.DataFrame = predictions_valid.describe(percentiles=[0.1, 0.01, 0.001])
            print(pred_descr)
            stat = Statistics.from_numpy(predictions[target].to_numpy())
            bin_size_multi = 0.2
            bins = get_symmetric_stddev_bins(stat, binsize_multi=bin_size_multi)
            binned_based_on_diff = []
            binned_based_on_diff_train = []
            for (start, end) in zip(bins[:-1], bins[1:]):
                binned_based_on_diff.append(predictions_valid.filter(
                    pl.col(target).is_between(start, end, closed='left')
                )[f"{target}_err_abs"].mean())
                # binned_based_on_diff_train.append(predictions_train.filter(
                #     pl.col(target).is_between(start, end, closed='left')
                # )["difficulty_err_abs"].mean())

            max_real_diff = predictions_valid[target].max()

            PlotData(
                x_ax=AxData(label='target'),
                y_ax=AxData(label='predicted'),
                save_path=f"{baseStr}_pred_valid.pdf",
                graphs=[
                    Function(identity, range_from=predictions[target].to_numpy(), label="target"),
                    Plot(yData=predictions_valid[f"predicted_{target}"].to_list(), xData=predictions_valid[target].to_list(), label="predicted", kind="scatter"),
                ]
            ).create().draw_and_save()

            PlotData(
                x_ax=AxData(label='target'),
                y_ax=AxData(label='predicted'),
                save_path=f"{baseStr}_pred_train.pdf",
                graphs=[
                    Function(identity, range_from=predictions[target].to_numpy(), label="target"),
                    Plot(yData=predictions_train[f"predicted_{target}"].to_list(), xData=predictions_train[target].to_list(), label="predicted", kind="scatter"),
                ]
            ).create().draw_and_save()
            # i need to pass an explicit dtype, otherwise - withthe latest data - it infers the wrong type. 
            binned_based_on_diff = np.array(binned_based_on_diff, dtype=np.float32)
            mae_binned = PlotData(
                title=f'Mean accross bins',
                x_ax=AxData(label="bin start value"),
                y_ax=AxData(label="mean absolute error"),
                save_path=f"{baseStr}_mae_binned.pdf",
                graphs=[
                    Bar(
                        # matplotlib doesnt like Nullish values, so we skip these
                        heights= np.select(~np.isnan(binned_based_on_diff), binned_based_on_diff),
                        xPositions= np.select(~np.isnan(binned_based_on_diff), bins[:-1]) + 0.5 * (1 - 0.8) * stat.std_dev * bin_size_multi,
                        width= 0.8 * stat.std_dev * bin_size_multi,
                    ),
                    VertLine(stat.mean, label="mean", linestyle= ":", color="green"),
                ]
            ).create().draw_and_save()
        except Exception as ex:
            print(ex)
            continue


    # per_category_train.create().draw_and_save()
    # per_category_validate.create().draw_and_save()