import sklearn.model_selection as skl_ms
import polars as pl
import numpy as np
import tensorflow as tf
from typing import Iterable, Callable

def data_split(
    data: pl.DataFrame,
    rdm_seed: int,
) -> (pl.DataFrame, pl.DataFrame, pl.DataFrame):
    """
    train, validation, test split
    """
    idx = pl.Series(name="selection_idxs", values=np.arange(len(data)))
    meta_train_idx, test_idx = skl_ms.train_test_split(idx, test_size=0.2, random_state=rdm_seed)
    train_idx, validate_idx = skl_ms.train_test_split(meta_train_idx, test_size=0.1, random_state=rdm_seed)
    
    in_train = idx.is_in(train_idx)
    in_validate = idx.is_in(validate_idx)
    in_test = idx.is_in(test_idx)

    return (data.filter(in_train), data.filter(in_validate), data.filter(in_test))

def to_tf_dataset(
    text_data: Iterable[str],
    target_data: Iterable[float],
    encoder: Callable[[Iterable[str]], Iterable[str]],
    batch_size: int,
):
    return (
        tf.data.Dataset
            # why slices?!
            .from_tensor_slices((
                encoder(text_data),
                target_data,
            ))
            .batch(batch_size)
            .prefetch(tf.data.experimental.AUTOTUNE)
    )

def prepare_datasets(
    train_df: pl.DataFrame,
    validate_df: pl.DataFrame,
    test_df: pl.DataFrame,
    input_text_key: str,
    target_key: str,
    encoder: Callable[[Iterable[str]], Iterable[str]],
    batch_size: int,
):
    def apply(df_to_convert: pl.DataFrame):
        return to_tf_dataset(
            df_to_convert[input_text_key],
            df_to_convert[target_key],
            encoder=encoder,
            batch_size=batch_size
        )

    return (
        apply(train_df), 
        apply(validate_df),
        apply(test_df),
    )

def load_and_split_data(
    csv_path: str,
    split_rdm_seed: int,
) -> (pl.DataFrame, pl.DataFrame, pl.DataFrame):
    "unique_entry_key = question text in my (smaier) application"
    data_df: pl.DataFrame = (pl
        .read_csv(csv_path)
        # DEV: limit to 100 entries
        # .limit(100)
    )
    (train_df, validate_df, test_df) = data_split(
        data_df,
        rdm_seed = split_rdm_seed,
    )
    print('#train', len(train_df))
    print('#validate', len(validate_df))
    print('#test', len(test_df))
    return (train_df, validate_df, test_df)
