from typing import Iterable
import polars as pl
import tensorflow as tf
import numpy as np
import dataclasses
import statistics
from config import *
import json
import train_strategy.train_test_validate_split as ttv_split
import model as model_helpers
import train_strategy.leave_one_out as loo
import pprint

def enem_training(seed: int):
    from itertools import product

    targets = [
        "irt_b_marinho_maxnorm", 
        "correctness",
    ]

    model_configs = [
        PresetConfig(
            model=ModelConfig("distilbert-base-multilingual-cased"), 
            target=target,
        )
        for target
        # You can put a full matrix of metaparameters to check in here with product(list1, list2) and tuple destructuring.
        # Beware, that this has of course massive influence on the time this job takes.
        # Also, tensorflow/keras has a "Memory Leak" of sorts.
        # It keeps old models around in (GPU) memory, when that is full they are moved between GPU and system RAM.
        # `reset_tf_memory` in model.py (called indirectly) is supposed to get rid of that,
        # however from my experience it doesn't actually.

        # If its discrete configs you want to test, do so with an explicit list instead
        in targets
    ]

    csv_path = './sources/enem_prepared.csv'

    input_text_key = 'question_and_answer_text'

    (
        train_df,
        validate_df,
        test_df,
    ) = ttv_split.load_and_split_data(
        csv_path = csv_path,
        split_rdm_seed = 1234,
    )

    for config in model_configs:
        config = AllConfig(model = config.model, training= config.training, seed=seed, target=config.target)
        print(f"""
###################### START
{pprint.pformat(config)}
######################""")

        def encoder(texts: Iterable[str]):
            "this captures some settings while offering an easy interface (aka its a closure)"
            return model_helpers.encode_using(
                texts,
                config.model.huggingface_name,
                config.model.context_length,
            )

        (
            train_tf_dataset, 
            validate_tf_dataset, 
            test_tf_dataset,
        ) = ttv_split.prepare_datasets(
            train_df, 
            validate_df, 
            test_df,
            input_text_key = input_text_key,
            target_key = config.target,
            encoder = encoder,
            batch_size = config.training.batch_size,
        )
        str_identifier = f"hash[{hex(hash(config))}]"

        with open(f"{str_identifier}_config.json", mode="x") as filehandle:
            json.dump(dataclasses.asdict(config), filehandle, indent="\t")

        model = model_helpers.SIDEEFFECTS_create_model_with_seed(config)

        train_steps = len(train_df) // config.training.batch_size

        history = model.fit(
            (train_tf_dataset
                .repeat(tf.data.experimental.INFINITE_CARDINALITY)
            ),
            validation_data=validate_tf_dataset,
            # DEV: switch to one epoch
            # epochs=60,
            epochs=config.training.epochs,
            # epochs=1,
            steps_per_epoch= train_steps, # floor operation
            callbacks = [
                tf.keras.callbacks.EarlyStopping(
                    # monitor='loss',
                    monitor='val_loss',
                    mode='min',
                    patience=config.training.earlystop_patience,
                    restore_best_weights=True,
                )
            ],
        )

        descriptors = model.metrics_names

        description = 'Train result:\n'
        train_result = model.evaluate(train_tf_dataset, verbose=0) #steps=train_steps, 
        description += str(list(zip(descriptors, train_result))) + '\n'

        description += 'Validation result:\n'
        validate_result = model.evaluate(validate_tf_dataset, verbose=0)
        description += str(list(zip(descriptors, validate_result))) + '\n'

        print(description)

        print(f"""
            SEED:               {seed}
            train    MSE (SD):  {sum(train_result   ):8} ({statistics.stdev(train_result   ):8})
            validate MSE (SD):  {sum(validate_result):8} ({statistics.stdev(validate_result):8})
        """)

        # AAANd we repeat the predictions again, after just doing them with model.evaluate...
        train_predictions: np.array
        train_predictions = model.predict(train_tf_dataset) #steps=train_steps
        validate_predictions = model.predict(validate_tf_dataset)
        test_predictions = model.predict(test_tf_dataset)

        model_results = pl.concat([
            train_df.with_columns([
                pl.lit("train").alias("split"),
                pl.Series(f"predicted_{config.target}", train_predictions.flatten()),
            ]),
            validate_df.with_columns([
                pl.lit("validate").alias("split"),
                pl.Series(f"predicted_{config.target}", validate_predictions.flatten()),
            ]),
            test_df.with_columns([
                pl.lit("test").alias("split"),
                pl.Series(f"predicted_{config.target}", test_predictions.flatten()),
            ]),
        ]).sort(pl.col("question_id"))

        model_results.write_csv(f"{str_identifier}_predictions.csv")

        pl.DataFrame(history.history).write_csv(f"{str_identifier}_history.csv")
