from typing import Iterable
import polars as pl
import tensorflow as tf
import numpy as np
import dataclasses
import statistics
from config import *
import json
import train_strategy.train_test_validate_split as ttv_split
import model as model_helpers
import train_strategy.leave_one_out as loo
import pprint

def hft_training(seed: int):
    # TODO:
    # csv_path = './sources/enem_prepared.csv'
    csv_path = './sources/hft_prepared.csv'

    input_text_key = 'question_and_answer_text'

    config = AllConfig(
        model = ModelConfig("distilbert-base-multilingual-cased"), 
        # TODO:
        training=TrainingConfig(training_kind="leave_one_out", epochs=100),
        # training=TrainingConfig(training_kind="leave_one_out"),
        # TODO:
        # target="irt_b_marinho_maxnorm",
        # target="correctness",
        target="mean_share_total_points",
        seed=seed,
    )

    str_identifier = f"hash[{hex(hash(config))}]"

    with open(f"{str_identifier}_config.json", mode="x") as filehandle:
        json.dump(dataclasses.asdict(config), filehandle, indent="\t")

    data = pl.read_csv(csv_path)
    # DEV: shorten data during pipeline tests
    # data = data[:2]

    input_encoding_key = f"{input_text_key}_encoded"
    data_with_enc: pl.DataFrame = data.with_columns(
        pl.Series(model_helpers.encode_using(
            data[input_text_key],
            config.model.huggingface_name, 
            config.model.context_length,
        ), dtype=pl.datatypes.List(pl.datatypes.Int32)).alias(input_encoding_key)
    )

    leave_one_out_pred: [float|None] = [None] * len(data_with_enc)
    "this shall contain the predictions of the left out datapoint of each respective round"
    leave_one_out_eval: [float|None] = [None] * len(data_with_enc)
    "this shall contain the RMS on the training data of each respective round. Not sure what I would actually use it for."
    # turns out that `model.metrics_names` doesnt actually contain anything
    # until the model was fitted for the first time.
    # That particular gotcha got me, though it is actually mentioned in the documentation!
    # leave_one_out_hist: pl.DataFrame = pl.DataFrame({
    #     **{
    #         fieldname: pl.Series() for fieldname in model.metrics_names
    #     },
    #     "left_out": pl.Series(),
    # })
    leave_one_out_hist: pl.DataFrame | None = None
    "this shall contain the history"

    for (train, (test_idx, test)) in loo.split_data_iter(data_with_enc):
        print(f"""
###################### START
Left out: {test_idx}/{len(data_with_enc)}
Config:
{pprint.pformat(config)}
######################""")
        model = model_helpers.SIDEEFFECTS_create_model_with_seed(config)

        assert len(test_idx) == 1
        # print(
        # train[input_encoding_key].to_list(),
        #     train[config.target].to_list(),)
        # dataset = tf.data.Dataset.from_tensor_slices((
        #     train[input_encoding_key].to_list(),
        #     train[config.target].to_list(),
        # ))
        # print(dataset)
        # DEV: Exit early during local dev. 
        # Exit with a nonzero exit code (=> failure) 
        # to hopefully get a Failure mail if I push that to the cluster by accident
        # exit(1)
        history = model.fit(
            np.array(train[input_encoding_key].to_list()),
            train[config.target].to_numpy(),
            batch_size=config.training.batch_size,
            epochs=config.training.epochs,
            callbacks = [
                tf.keras.callbacks.EarlyStopping(
                    monitor='loss',
                    # actually unavailable here, as we dont have validation data
                    # monitor='val_loss',
                    mode='min',
                    patience=config.training.earlystop_patience,
                    restore_best_weights=True,
                ),
            ],
        )

        new_hist = pl.DataFrame(
            history.history
        ).with_columns(pl.lit(test_idx[0]).alias(("left_out")))
        if leave_one_out_hist is None:
            leave_one_out_hist = new_hist
        else:
            leave_one_out_hist = pl.concat([
                leave_one_out_hist,
                new_hist,
            ])

        eval_res = model.evaluate(
            np.array(train[input_encoding_key].to_list()),
            train[config.target].to_numpy(),
            verbose=0
        )

        rms_idx = 2

        assert model.metrics_names[rms_idx] == "root_mean_squared_error"

        leave_one_out_eval[test_idx[0]] = eval_res[rms_idx]

        leave_one_out_pred[test_idx[0]] = model.predict(
            np.array(test[input_encoding_key].to_list()),
        )[0][0] # IDK why i need to index twice.

    assert_no_None(leave_one_out_eval)
    assert_no_None(leave_one_out_pred)

    print(f"""
        SEED:           {seed}
        leave-one-out   MSE (SD):  {np.mean(leave_one_out_eval):8} ({statistics.stdev(leave_one_out_eval):8})
    """)

    data_with_predict: pl.DataFrame = data.with_columns(
        pl.Series(leave_one_out_pred).alias(f"predicted_{config.target}"),
    )

    data_with_predict.write_csv(f"{str_identifier}_predictions.csv")

    leave_one_out_hist.write_csv(f"{str_identifier}_history.csv")

def assert_no_None(arg: list):
    for itm in arg:
        # numpy fix
        # this ordering of not and instance check also allows for numpy arrays as elements,
        # with `itm is not None` it fails due to an (IMO) overzelous check by numpy 
        # (together with a very confusing error message for the source code)
        assert not itm is None
